@@ -2,10 +2,12 @@
 pub mod defaults;
 pub mod term;
 
+use defaults::*;
 use regex::Regex;
 use serde::Serialize;
 use serde::{de, Deserialize};
 use std::num::ParseIntError;
+use std::ops::Mul;
 
 pub type ColorWGPU = wgpu::Color;
 pub type ColorArray = [f32; 4];
@@ -18,6 +20,32 @@ pub struct ColorRgb {
     pub b: u8,
 }
 
+impl Mul<f32> for ColorRgb {
+    type Output = ColorRgb;
+
+    fn mul(self, rhs: f32) -> ColorRgb {
+        let result = ColorRgb {
+            r: (f32::from(self.r) * rhs).clamp(0.0, 255.0) as u8,
+            g: (f32::from(self.g) * rhs).clamp(0.0, 255.0) as u8,
+            b: (f32::from(self.b) * rhs).clamp(0.0, 255.0) as u8,
+        };
+
+        tracing::trace!(
+            "Scaling ColorRgb by {} from {:?} to {:?}",
+            rhs,
+            self,
+            result
+        );
+        result
+    }
+}
+
+impl From<&ColorRgb> for ColorArray {
+    fn from(color: &ColorRgb) -> ColorArray {
+        color.to_arr()
+    }
+}
+
 impl ColorRgb {
     pub fn from_color_arr(arr: ColorArray) -> ColorRgb {
         ColorRgb {
@@ -121,109 +149,109 @@ pub struct Colors {
     #[serde(default = "defaults::white", deserialize_with = "deserialize_to_arr")]
     pub white: ColorArray,
     #[serde(
-        default = "defaults::dim_black",
-        deserialize_with = "deserialize_to_arr",
+        default = "Option::default",
+        deserialize_with = "deserialize_to_arr_opt",
         rename = "dim-black"
     )]
-    pub dim_black: ColorArray,
+    pub dim_black: Option<ColorArray>,
     #[serde(
-        default = "defaults::dim_blue",
-        deserialize_with = "deserialize_to_arr",
+        default = "Option::default",
+        deserialize_with = "deserialize_to_arr_opt",
         rename = "dim-blue"
     )]
-    pub dim_blue: ColorArray,
+    pub dim_blue: Option<ColorArray>,
     #[serde(
-        default = "defaults::dim_cyan",
-        deserialize_with = "deserialize_to_arr",
+        default = "Option::default",
+        deserialize_with = "deserialize_to_arr_opt",
         rename = "dim-cyan"
     )]
-    pub dim_cyan: ColorArray,
+    pub dim_cyan: Option<ColorArray>,
     #[serde(
-        default = "defaults::dim_foreground",
-        deserialize_with = "deserialize_to_arr",
+        default = "Option::default",
+        deserialize_with = "deserialize_to_arr_opt",
         rename = "dim-foreground"
     )]
-    pub dim_foreground: ColorArray,
+    pub dim_foreground: Option<ColorArray>,
     #[serde(
-        default = "defaults::dim_green",
-        deserialize_with = "deserialize_to_arr",
+        default = "Option::default",
+        deserialize_with = "deserialize_to_arr_opt",
         rename = "dim-green"
     )]
-    pub dim_green: ColorArray,
+    pub dim_green: Option<ColorArray>,
     #[serde(
-        default = "defaults::dim_magenta",
-        deserialize_with = "deserialize_to_arr",
+        default = "Option::default",
+        deserialize_with = "deserialize_to_arr_opt",
         rename = "dim-magenta"
     )]
-    pub dim_magenta: ColorArray,
+    pub dim_magenta: Option<ColorArray>,
     #[serde(
-        default = "defaults::dim_red",
-        deserialize_with = "deserialize_to_arr",
+        default = "Option::default",
+        deserialize_with = "deserialize_to_arr_opt",
         rename = "dim-red"
     )]
-    pub dim_red: ColorArray,
+    pub dim_red: Option<ColorArray>,
     #[serde(
-        default = "defaults::dim_white",
-        deserialize_with = "deserialize_to_arr",
+        default = "Option::default",
+        deserialize_with = "deserialize_to_arr_opt",
         rename = "dim-white"
     )]
-    pub dim_white: ColorArray,
+    pub dim_white: Option<ColorArray>,
     #[serde(
-        default = "defaults::dim_yellow",
-        deserialize_with = "deserialize_to_arr",
+        default = "Option::default",
+        deserialize_with = "deserialize_to_arr_opt",
         rename = "dim-yellow"
     )]
-    pub dim_yellow: ColorArray,
+    pub dim_yellow: Option<ColorArray>,
     #[serde(
-        default = "defaults::light_black",
+        default = "default_light_black",
         deserialize_with = "deserialize_to_arr",
         rename = "light-black"
     )]
     pub light_black: ColorArray,
     #[serde(
-        default = "defaults::light_blue",
+        default = "default_light_blue",
         deserialize_with = "deserialize_to_arr",
         rename = "light-blue"
     )]
     pub light_blue: ColorArray,
     #[serde(
-        default = "defaults::light_cyan",
+        default = "default_light_cyan",
         deserialize_with = "deserialize_to_arr",
         rename = "light-cyan"
     )]
     pub light_cyan: ColorArray,
     #[serde(
-        default = "defaults::light_foreground",
-        deserialize_with = "deserialize_to_arr",
+        default = "Option::default",
+        deserialize_with = "deserialize_to_arr_opt",
         rename = "light-foreground"
     )]
-    pub light_foreground: ColorArray,
+    pub light_foreground: Option<ColorArray>,
     #[serde(
-        default = "defaults::light_green",
+        default = "default_light_green",
         deserialize_with = "deserialize_to_arr",
         rename = "light-green"
     )]
     pub light_green: ColorArray,
     #[serde(
-        default = "defaults::light_magenta",
+        default = "default_light_magenta",
         deserialize_with = "deserialize_to_arr",
         rename = "light-magenta"
     )]
     pub light_magenta: ColorArray,
     #[serde(
-        default = "defaults::light_red",
+        default = "default_light_red",
         deserialize_with = "deserialize_to_arr",
         rename = "light-red"
     )]
     pub light_red: ColorArray,
     #[serde(
-        default = "defaults::light_white",
+        default = "default_light_white",
         deserialize_with = "deserialize_to_arr",
         rename = "light-white"
     )]
     pub light_white: ColorArray,
     #[serde(
-        default = "defaults::light_yellow",
+        default = "default_light_yellow",
         deserialize_with = "deserialize_to_arr",
         rename = "light-yellow"
     )]
@@ -290,24 +318,24 @@ impl Default for Colors {
             cyan: defaults::cyan(),
             magenta: defaults::magenta(),
             white: defaults::white(),
-            dim_black: defaults::dim_black(),
-            dim_blue: defaults::dim_blue(),
-            dim_cyan: defaults::dim_cyan(),
-            dim_foreground: defaults::dim_foreground(),
-            dim_green: defaults::dim_green(),
-            dim_magenta: defaults::dim_magenta(),
-            dim_red: defaults::dim_red(),
-            dim_white: defaults::dim_white(),
-            dim_yellow: defaults::dim_yellow(),
-            light_black: defaults::light_black(),
-            light_blue: defaults::light_blue(),
-            light_cyan: defaults::light_cyan(),
-            light_foreground: defaults::light_foreground(),
-            light_green: defaults::light_green(),
-            light_magenta: defaults::light_magenta(),
-            light_red: defaults::light_red(),
-            light_white: defaults::light_white(),
-            light_yellow: defaults::light_yellow(),
+            dim_black: None,
+            dim_blue: None,
+            dim_cyan: None,
+            dim_foreground: None,
+            dim_green: None,
+            dim_magenta: None,
+            dim_red: None,
+            dim_white: None,
+            dim_yellow: None,
+            light_black: default_light_black(),
+            light_blue: default_light_blue(),
+            light_cyan: default_light_cyan(),
+            light_foreground: None,
+            light_green: default_light_green(),
+            light_magenta: default_light_magenta(),
+            light_red: default_light_red(),
+            light_white: default_light_white(),
+            light_yellow: default_light_yellow(),
             selection_background: defaults::selection_background(),
             selection_foreground: defaults::selection_foreground(),
             search_match_background: defaults::search_match_background(),
@@ -633,6 +661,19 @@ where
     }
 }
 
+pub fn deserialize_to_arr_opt<'de, D>(
+    deserializer: D,
+) -> Result<Option<ColorArray>, D::Error>
+where
+    D: de::Deserializer<'de>,
+{
+    let s = String::deserialize(deserializer)?;
+    match ColorBuilder::from_hex(s, Format::SRGB0_1) {
+        Ok(color) => Ok(Some(color.to_arr())),
+        Err(e) => Err(serde::de::Error::custom(e)),
+    }
+}
+
 #[cfg(test)]
 mod tests {
     use super::*;
