@@ -22,7 +22,11 @@ pub struct Atlas {
 }
 
 impl Atlas {
-    pub fn new(device: &wgpu::Device, backend: wgpu::Backend) -> Self {
+    pub fn new(
+        device: &wgpu::Device,
+        backend: wgpu::Backend,
+        context: &crate::context::Context,
+    ) -> Self {
         let layers = match backend {
             // On the GL backend we start with 2 layers, to help wgpu figure
             // out that this texture is `GL_TEXTURE_2D_ARRAY` rather than `GL_TEXTURE_2D`
@@ -37,13 +41,15 @@ impl Atlas {
             depth_or_array_layers: layers.len() as u32,
         };
 
+        let texture_format = context.get_optimal_texture_format(4); // RGBA
+
         let texture = device.create_texture(&wgpu::TextureDescriptor {
             label: Some("image texture atlas"),
             size: extent,
             mip_level_count: 1,
             sample_count: 1,
             dimension: wgpu::TextureDimension::D2,
-            format: wgpu::TextureFormat::Rgba8Unorm,
+            format: texture_format,
             // with gamma correction
             // wgpu::TextureFormat::Rgba8UnormSrgb
             // } else {
@@ -75,6 +81,48 @@ impl Atlas {
         self.layers.len()
     }
 
+    pub fn clear(
+        &mut self,
+        device: &wgpu::Device,
+        backend: wgpu::Backend,
+        context: &crate::context::Context,
+    ) {
+        // Reset all layers to empty
+        self.layers = match backend {
+            wgpu::Backend::Gl => vec![Layer::Empty, Layer::Empty],
+            _ => vec![Layer::Empty],
+        };
+
+        // Recreate the texture with the reset layers
+        let extent = wgpu::Extent3d {
+            width: SIZE,
+            height: SIZE,
+            depth_or_array_layers: self.layers.len() as u32,
+        };
+
+        let texture_format = context.get_optimal_texture_format(4); // RGBA
+
+        self.texture = device.create_texture(&wgpu::TextureDescriptor {
+            label: Some("image texture atlas"),
+            size: extent,
+            mip_level_count: 1,
+            sample_count: 1,
+            dimension: wgpu::TextureDimension::D2,
+            format: texture_format,
+            usage: wgpu::TextureUsages::COPY_DST
+                | wgpu::TextureUsages::COPY_SRC
+                | wgpu::TextureUsages::TEXTURE_BINDING,
+            view_formats: &[],
+        });
+
+        self.texture_view = self.texture.create_view(&wgpu::TextureViewDescriptor {
+            dimension: Some(wgpu::TextureViewDimension::D2Array),
+            ..Default::default()
+        });
+
+        tracing::info!("Layer atlas cleared");
+    }
+
     pub fn upload(
         &mut self,
         device: &wgpu::Device,
@@ -313,15 +361,15 @@ impl Atlas {
         });
 
         encoder.copy_buffer_to_texture(
-            wgpu::ImageCopyBuffer {
+            wgpu::TexelCopyBufferInfo {
                 buffer: &buffer,
-                layout: wgpu::ImageDataLayout {
+                layout: wgpu::TexelCopyBufferLayout {
                     offset: offset as u64,
                     bytes_per_row: Some(4 * image_dimensions.0 + padding),
                     rows_per_image: Some(image_dimensions.1),
                 },
             },
-            wgpu::ImageCopyTexture {
+            wgpu::TexelCopyTextureInfo {
                 texture: &self.texture,
                 mip_level: 0,
                 origin: wgpu::Origin3d {
@@ -375,7 +423,7 @@ impl Atlas {
             }
 
             encoder.copy_texture_to_texture(
-                wgpu::ImageCopyTexture {
+                wgpu::TexelCopyTextureInfo {
                     texture: &self.texture,
                     mip_level: 0,
                     origin: wgpu::Origin3d {
@@ -385,7 +433,7 @@ impl Atlas {
                     },
                     aspect: wgpu::TextureAspect::default(),
                 },
-                wgpu::ImageCopyTexture {
+                wgpu::TexelCopyTextureInfo {
                     texture: &new_texture,
                     mip_level: 0,
                     origin: wgpu::Origin3d {
