@@ -170,7 +170,7 @@ impl LayerBrush {
                     binding: 0,
                     visibility: wgpu::ShaderStages::FRAGMENT,
                     ty: wgpu::BindingType::Texture {
-                        sample_type: wgpu::TextureSampleType::Float { filterable: true },
+                        sample_type: context.get_optimal_texture_sample_type(),
                         view_dimension: wgpu::TextureViewDimension::D2Array,
                         multisampled: false,
                     },
@@ -184,11 +184,15 @@ impl LayerBrush {
             bind_group_layouts: &[&constant_layout, &texture_layout],
         });
 
+        let shader_source = if context.supports_f16() {
+            include_str!("image.wgsl")
+        } else {
+            include_str!("image_f32.wgsl")
+        };
+
         let shader = device.create_shader_module(wgpu::ShaderModuleDescriptor {
             label: Some("layer image shader"),
-            source: wgpu::ShaderSource::Wgsl(std::borrow::Cow::Borrowed(include_str!(
-                "image.wgsl"
-            ))),
+            source: wgpu::ShaderSource::Wgsl(std::borrow::Cow::Borrowed(shader_source)),
         });
 
         let pipeline = device.create_render_pipeline(&wgpu::RenderPipelineDescriptor {
@@ -198,7 +202,7 @@ impl LayerBrush {
             vertex: wgpu::VertexState {
                 compilation_options: wgpu::PipelineCompilationOptions::default(),
                 module: &shader,
-                entry_point: "vs_main",
+                entry_point: Some("vs_main"),
                 buffers: &[
                     wgpu::VertexBufferLayout {
                         array_stride: mem::size_of::<Vertex>() as u64,
@@ -225,7 +229,7 @@ impl LayerBrush {
             fragment: Some(wgpu::FragmentState {
                 compilation_options: wgpu::PipelineCompilationOptions::default(),
                 module: &shader,
-                entry_point: "fs_main",
+                entry_point: Some("fs_main"),
                 targets: &[Some(wgpu::ColorTargetState {
                     format: context.format,
                     blend: Some(wgpu::BlendState {
@@ -269,7 +273,7 @@ impl LayerBrush {
             usage: wgpu::BufferUsages::INDEX,
         });
 
-        let texture_atlas = Atlas::new(device);
+        let texture_atlas = Atlas::new(device, context.adapter_info.backend, context);
 
         let texture = device.create_bind_group(&wgpu::BindGroupDescriptor {
             label: Some("image texture atlas bind group"),
@@ -341,7 +345,7 @@ impl LayerBrush {
         let texture_version = self.texture_atlas.layer_count();
 
         if self.texture_version != texture_version {
-            log::info!("Atlas has grown. Recreating bind group...");
+            tracing::info!("Atlas has grown. Recreating bind group...");
 
             self.texture = device.create_bind_group(&wgpu::BindGroupDescriptor {
                 label: Some("image texture atlas bind group"),
@@ -401,7 +405,7 @@ impl LayerBrush {
         let texture_version = self.texture_atlas.layer_count();
 
         if self.texture_version != texture_version {
-            log::info!("Atlas has grown. Recreating bind group...");
+            tracing::info!("Atlas has grown. Recreating bind group...");
 
             self.texture = device.create_bind_group(&wgpu::BindGroupDescriptor {
                 label: Some("image texture atlas bind group"),
@@ -502,6 +506,29 @@ impl LayerBrush {
         }
     }
 
+    pub fn clear_atlas(
+        &mut self,
+        device: &wgpu::Device,
+        backend: wgpu::Backend,
+        context: &crate::context::Context,
+    ) {
+        self.texture_atlas.clear(device, backend, context);
+        self.texture_version = self.texture_atlas.layer_count();
+
+        // Recreate the bind group with the new atlas
+        self.texture = device.create_bind_group(&wgpu::BindGroupDescriptor {
+            label: Some("image texture atlas bind group"),
+            layout: &self.texture_layout,
+            entries: &[wgpu::BindGroupEntry {
+                binding: 0,
+                resource: wgpu::BindingResource::TextureView(self.texture_atlas.view()),
+            }],
+        });
+
+        // Clear the raster cache as well
+        self.raster_cache.borrow_mut().clear();
+    }
+
     pub fn end_frame(&mut self) {
         self.raster_cache.borrow_mut().trim(&mut self.texture_atlas);
 
